"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.UserEventType = void 0;
const api_service_1 = require("./api.service");
const storage_1 = require("./storage");
const featureFlagResource = '/feature-flags';
const userResource = '/users';
const userEventResource = '/user-events';
const storagePath = '@growthflags/js-sdk';
const consoleLogPrefix = '@growthflags/js-sdk error:';
var UserEventType;
(function (UserEventType) {
    UserEventType["FeatureViewed"] = "featureViewed";
})(UserEventType = exports.UserEventType || (exports.UserEventType = {}));
class FeatureFlags {
    constructor({ publicApiKey, env }) {
        this._apiKey = publicApiKey;
        this._env = env;
        this._features = {};
        this._featureOverrides = [];
        this._configs = {};
    }
    fetchFeatureFlags(user) {
        return __awaiter(this, void 0, void 0, function* () {
            if ((user === null || user === void 0 ? void 0 : user.email) || (user === null || user === void 0 ? void 0 : user.id)) {
                yield this._createUser({
                    id: user.id,
                    email: user.email,
                    data: user.data
                });
            }
            const storageData = this._getFromStorage();
            if (storageData) {
                this._features = storageData.features || {};
                this._configs = storageData.configs || {};
            }
            return this._fetchFlags();
        });
    }
    isOn(featureName) {
        return Boolean(this._features[featureName]);
    }
    mergeFeatures(features) {
        let newFeatures = features;
        this._featureOverrides.forEach((featureOverride) => {
            newFeatures[featureOverride.name] = featureOverride.enabled;
        });
        return newFeatures;
    }
    setFeatures(featureOverrides) {
        this._featureOverrides = featureOverrides;
        this._features = this.mergeFeatures(this._features);
    }
    getFeature(featureName, options = {}) {
        return {
            name: featureName,
            enabled: this.isOn(featureName),
            config: this.getConfig(featureName, options.defaultConfig),
        };
    }
    getConfig(featureName, defaultConfig) {
        if (this._configs[featureName]) {
            return JSON.parse(this._configs[featureName]);
        }
        return defaultConfig;
    }
    trackFeatureView(featureName) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            if (!this._user)
                return;
            const data = {
                userId: (_a = this._user) === null || _a === void 0 ? void 0 : _a._id,
                type: UserEventType.FeatureViewed,
                data: {
                    featureName
                }
            };
            const config = {
                headers: {
                    Authorization: 'Bearer ' + this._apiKey,
                },
            };
            try {
                yield api_service_1.default.post(`${userEventResource}`, data, config);
            }
            catch (error) {
                console.log(consoleLogPrefix, error);
            }
        });
    }
    _fetchFlags() {
        return __awaiter(this, void 0, void 0, function* () {
            const params = { env: this._env };
            if (this._user) {
                params.userId = this._user._id;
            }
            const config = {
                headers: {
                    Authorization: 'Bearer ' + this._apiKey,
                },
            };
            try {
                const response = yield api_service_1.default.get(`${featureFlagResource}/features`, params, config);
                const features = response.features || {};
                this._features = this.mergeFeatures(features);
                this._configs = response.configs || {};
                this._saveToStorage(response);
                return response;
            }
            catch (error) {
                console.log(consoleLogPrefix, error);
            }
        });
    }
    _createUser(user) {
        return __awaiter(this, void 0, void 0, function* () {
            const { id, email, data } = user;
            const reqData = { env: this._env, id, email, data };
            const config = {
                headers: {
                    Authorization: 'Bearer ' + this._apiKey,
                },
            };
            try {
                const response = yield api_service_1.default.post(`${userResource}`, reqData, config);
                this._user = response;
                this._saveToStorage({ user: response });
            }
            catch (error) {
                console.log(consoleLogPrefix, error);
            }
        });
    }
    _saveToStorage(data) {
        if (!storage_1.default) {
            return;
        }
        try {
            const localStorageData = this._getFromStorage();
            const updatedData = localStorageData ? Object.assign(Object.assign({}, localStorageData), data) : Object.assign({}, data);
            const stringData = JSON.stringify(updatedData);
            storage_1.default.setItem(`${storagePath}`, stringData);
        }
        catch (error) {
            console.log(consoleLogPrefix, error);
        }
    }
    _getFromStorage() {
        if (!storage_1.default) {
            return null;
        }
        try {
            const localStorageData = storage_1.default.getItem(`${storagePath}`);
            return localStorageData ? JSON.parse(localStorageData) : null;
        }
        catch (error) {
            console.log(consoleLogPrefix, error);
        }
    }
}
let instance;
exports.default = {
    create: ({ publicApiKey, env }) => {
        if (!publicApiKey) {
            throw new RangeError('Invalid arguments: "publicApiKey" must be provided.');
        }
        if (!env) {
            throw new RangeError('Invalid arguments: "env" must be provided.');
        }
        if (instance) {
            return instance;
        }
        instance = new FeatureFlags({ publicApiKey, env });
        return instance;
    }
};
